#ifdef NDEBUG
#undef NDEBUG
#endif

#include <algorithm>
#include <cmath>
#include <cassert>
#include <iostream>

#include "config.hh"
#include "EventGenerator.hh"

#include "HEJ/Event.hh"
#include "HEJ/MatrixElement.hh"
#include "HEJ/Particle.hh"
#include "HEJ/PDF.hh"
#include "HEJ/Ranlux64.hh"
#include "HEJ/utility.hh"

using namespace HEJFOG;

bool pass_dR_cut(
    std::vector<fastjet::PseudoJet> const & jets,
    std::vector<HEJ::Particle> const & photons
){
  constexpr double delta_R_min = 0.7;
  for(auto const & jet: jets){
    for(auto const & photon: photons){
      if(jet.delta_R(photon.p) < delta_R_min) return false;
    }
  }
  return true;
}

int main(){
  constexpr double invGeV2_to_pb = 389379292.;
  constexpr double xs_ref = 0.00429198; // +- 1.0488e-05
  //calculated with HEJ revision 9570e3809613272ac4b8bf3236279ba23cf64d20

  auto config = load_config("config_h_2j_decay.yml");

  HEJ::Ranlux64 ran{};
  HEJFOG::EventGenerator generator{
    config.process,
    config.beam,
    HEJ::ScaleGenerator{
      config.scales.base,
      config.scales.factors,
      config.scales.max_ratio
    },
    config.jets,
    config.pdf_id,
    config.subleading_fraction,
    config.subleading_channels,
    config.particles_properties,
    config.Higgs_coupling,
    ran
  };

  double xs = 0., xs_err = 0.;
  for (int trials = 0; trials < config.events; ++trials){
    auto ev = generator.gen_event();
    if(generator.status() != good) continue;
    assert(ev.decays().size() == 1);
    const auto decay = begin(ev.decays());
    assert(ev.outgoing().size() > decay->first);
    const auto & the_Higgs = ev.outgoing()[decay->first];
    assert(the_Higgs.type == HEJ::pid::Higgs);
    assert(decay->second.size() == 2);
    auto const & gamma = decay->second;
    assert(gamma[0].type == HEJ::pid::photon);
    assert(gamma[1].type == HEJ::pid::photon);
    assert(HEJ::nearby_ep(gamma[0].p + gamma[1].p, the_Higgs.p, 1e-6));
    if(!pass_dR_cut(ev.jets(), gamma)) continue;
    ev.central().weight *= invGeV2_to_pb;
    ev.central().weight /= config.events;

    xs += ev.central().weight;
    xs_err += ev.central().weight*ev.central().weight;
  }
  xs_err = std::sqrt(xs_err);
  std::cout << xs_ref << " ~ " << xs << " +- " << xs_err << std::endl;

  assert(std::abs(xs - xs_ref) < 3*xs_err);
  assert(xs_err < 0.012*xs);
}
