/**
 * \file
 * \brief Declaration of the JetSplitter class
 *
 *  \authors   Jeppe Andersen, Tuomas Hapola, Marian Heil, Andreas Maier, Jennifer Smillie
 *  \date      2019
 *  \copyright GPLv2 or later
 */

#pragma once

#include <functional>
#include <vector>

#include "fastjet/JetDefinition.hh"

#include "HEJ/RNG.hh"

namespace fastjet {
  class PseudoJet;
}

namespace HEJ {
  //! Class to split jets into their constituents
  class JetSplitter {

  public:
    struct SplitResult {
      std::vector<fastjet::PseudoJet> constituents;
      double weight;
    };

    //! Constructor
    /**
     *  @param jet_def   Jet definition
     *  @param min_pt    Minimum jet transverse momentum
     *  @param ran       Random number generator
     */
    JetSplitter(
        fastjet::JetDefinition jet_def, double min_pt,
        HEJ::RNG & ran
    ):
      R_{jet_def.R()},
      min_jet_pt_{min_pt},
      jet_def_{jet_def},
      ran_{ran}
    {}

    //! Split a get into constituents
    /**
     *  @param j2split      Jet to be split
     *  @param ncons        Number of constituents
     *  @returns            The constituent momenta
     *                      together with the associated weight
     */
    SplitResult split(fastjet::PseudoJet const & j2split, int ncons) const;

    //! Maximum distance of constituents to jet axis
    static constexpr double R_factor = 5./3.;
  private:
    SplitResult Split2(fastjet::PseudoJet const & j2split) const;
    double sample_distance_2p(double & wt) const;

    double R_;
    double min_jet_pt_;
    fastjet::JetDefinition jet_def_;
    std::reference_wrapper<HEJ::RNG> ran_;
  };
}
