/** \file
 *  \brief Contains the MatrixElement Class
 *
 *  \authors   Jeppe Andersen, Tuomas Hapola, Marian Heil, Andreas Maier, Jennifer Smillie
 *  \date      2019
 *  \copyright GPLv2 or later
 */
#pragma once

#include <functional>
#include <vector>

#include "fastjet/PseudoJet.hh"

#include "HEJ/PDG_codes.hh"
#include "HEJ/Weights.hh"
#include "HEJ/config.hh"

namespace CLHEP {
  class HepLorentzVector;
}

namespace HEJ{
  class Event;
  class Particle;

  //! Class to calculate the squares of matrix elements
  class MatrixElement{
  public:
    /** \brief MatrixElement Constructor
     * @param alpha_s        Function taking the renormalisation scale
     *                       and returning the strong coupling constant
     * @param conf           General matrix element settings
     */
    MatrixElement(
        std::function<double (double)> alpha_s,
        MatrixElementConfig conf
    );

  /**
   * \brief squares of regulated HEJ matrix elements
   * @param event          The event for which to calculate matrix elements
   * @returns              The squares of HEJ matrix elements including virtual corrections
   *
   * cf. eq. (22) in \cite Andersen:2011hs
   *
   * \internal Relation to standard HEJ Met2: MatrixElement = Met2*shat^2/(pdfta*pdftb)
   */
    Weights operator()(Event const & event) const;

  //! Squares of HEJ tree-level matrix elements
  /**
   * @param event          The event for which to calculate matrix elements
   * @returns              The squares of HEJ matrix elements without virtual corrections
   *
   * cf. eq. (22) in \cite Andersen:2011hs
   */
    Weights tree(Event const & event) const;

   /**
    * \brief Virtual corrections to matrix element squares
   * @param event          The event for which to calculate matrix elements
    * @returns             The virtual corrections to the squares of the matrix elements
    *
    * The all order virtual corrections to LL in the MRK limit is
    * given by replacing 1/t in the scattering amplitude according to the
    * lipatov ansatz.
    *
    * cf. second-to-last line of eq. (22) in \cite Andersen:2011hs
    * note that indices are off by one, i.e. out[0].p corresponds to p_1
    */
    Weights virtual_corrections(Event const & event) const;

  private:

    Weights tree_param(
        Event const & event
    ) const;

    double tree_kin(Event const & event) const;

    double tree_param(
        Event const & event,
        double mur
    ) const;

    double virtual_corrections(
        Event const & event,
        double mur
    ) const;

    //! \internal cf. last line of eq. (22) in \cite Andersen:2011hs
    double omega0(
        double alpha_s, double mur,
        fastjet::PseudoJet const & q_j, double lambda
    ) const;

    double tree_kin_jets(
        Event const & ev
    ) const;
    double tree_kin_Higgs(
        Event const & ev
    ) const;
    double tree_kin_Higgs_first(
        Event const & ev
    ) const;
    double tree_kin_Higgs_last(
        Event const & ev
    ) const;

    /**
     * \internal
     * \brief Higgs inbetween extremal partons.
     *
     * Note that in the case of unordered emission, the Higgs is *always*
     * treated as if in between the extremal (FKL) partons, even if its
     * rapidity is outside the extremal parton rapidities
     */
    double tree_kin_Higgs_between(
        Event const & ev
    ) const;


    double tree_param_partons(
        double alpha_s, double mur,
        std::vector<Particle> const & partons
    ) const;


    std::vector<int> in_extremal_jet_indices(
        std::vector<fastjet::PseudoJet> const & partons
    ) const;


    std::vector<Particle> tag_extremal_jet_partons(
        Event const & ev
    ) const;

    double MH2_forwardH(
        CLHEP::HepLorentzVector p1out, CLHEP::HepLorentzVector p1in,
        pid::ParticleID type2,
        CLHEP::HepLorentzVector p2out, CLHEP::HepLorentzVector p2in,
        CLHEP::HepLorentzVector pH,
        double t1, double t2
    ) const;

    std::function<double (double)> alpha_s_;

    MatrixElementConfig param_;
  };


}
