/** \file
 *  \brief Contains the PhaseSpacePoint Class
 *
 *  \authors   Jeppe Andersen, Tuomas Hapola, Marian Heil, Andreas Maier, Jennifer Smillie
 *  \date      2019
 *  \copyright GPLv2 or later
 */

#pragma once

#include <array>
#include <functional>
#include <unordered_map>
#include <vector>

#include "HEJ/config.hh"
#include "HEJ/Particle.hh"
#include "HEJ/RNG.hh"

namespace HEJ{
  class Event;

  //! A point in resummation phase space
  class PhaseSpacePoint{
  public:
    //! Default PhaseSpacePoint Constructor
    PhaseSpacePoint() = default;

    //! PhaseSpacePoint Constructor
    /**
     * @param ev               Clustered Jet Event
     * @param conf             Configuration parameters
     * @param ran              Random number generator
     */
    PhaseSpacePoint(
        Event const & ev,
        PhaseSpacePointConfig conf,
        HEJ::RNG & ran
    );

    //! Get phase space point weight
    double weight() const{
      return weight_;
    }

    //! Access incoming particles
    std::array<Particle, 2> const & incoming() const{
      return incoming_;
    }

    //! Access outgoing particles
    std::vector<Particle> const & outgoing() const{
      return outgoing_;
    }


    //! Particle decays
    /**
     *  The key in the returned map corresponds to the index in the
     *  vector returned by outgoing()
     */
    std::unordered_map<size_t, std::vector<Particle>> const &  decays() const{
      return decays_;
    }

    static constexpr int ng_max = 1000;    //< maximum number of extra gluons

  private:

    std::vector<fastjet::PseudoJet> cluster_jets(
        std::vector<fastjet::PseudoJet> const & partons
    ) const;
    bool pass_resummation_cuts(
        std::vector<fastjet::PseudoJet> const & jets
    ) const;
    bool pass_extremal_cuts(
        fastjet::PseudoJet const & ext_parton,
        fastjet::PseudoJet const & jet
    ) const;
    int sample_ng(std::vector<fastjet::PseudoJet> const & Born_jets);
    int sample_ng_jets(int ng, std::vector<fastjet::PseudoJet> const & Born_jets);
    double probability_in_jet(
        std::vector<fastjet::PseudoJet> const & Born_jets
    ) const;
    std::vector<fastjet::PseudoJet> gen_non_jet(
        int ng_non_jet,
        double ptmin, double ptmax
    );
    void rescale_rapidities(
      std::vector<fastjet::PseudoJet> & partons,
      double ymin, double ymax
    );
    std::vector<fastjet::PseudoJet> reshuffle(
        std::vector<fastjet::PseudoJet> const & Born_jets,
        fastjet::PseudoJet const & q
    );
    bool jets_ok(
        std::vector<fastjet::PseudoJet> const & Born_jets,
        std::vector<fastjet::PseudoJet> const & partons
    ) const;
    void reconstruct_incoming(std::array<Particle, 2> const & Born_incoming);
    double phase_space_normalisation(
        int num_Born_jets,
        int num_res_partons
    ) const;
    std::vector<fastjet::PseudoJet> split(
        std::vector<fastjet::PseudoJet> const & jets, int ng_jets
    );
    std::vector<int> distribute_jet_partons(
        int ng_jets, std::vector<fastjet::PseudoJet> const & jets
    );
    std::vector<fastjet::PseudoJet> split(
        std::vector<fastjet::PseudoJet> const & jets,
        std::vector<int> const & np_in_jet
    );
    bool split_preserved_jets(
        std::vector<fastjet::PseudoJet> const & jets,
        std::vector<fastjet::PseudoJet> const & jet_partons
    ) const;
    template<class Particle>
    Particle const & most_backward_FKL(
        std::vector<Particle> const & partons
    ) const;
    template<class Particle>
    Particle const & most_forward_FKL(
        std::vector<Particle> const & partons
    ) const;
    template<class Particle>
    Particle & most_backward_FKL(std::vector<Particle> & partons) const;
    template<class Particle>
    Particle & most_forward_FKL(std::vector<Particle> & partons) const;
    bool extremal_ok(
        std::vector<fastjet::PseudoJet> const & partons
    ) const;
    void copy_AWZH_boson_from(Event const & event);

    bool momentum_conserved() const;

    bool unob_, unof_;

    double weight_;

    PhaseSpacePointConfig param_;


    std::array<Particle, 2> incoming_;
    std::vector<Particle> outgoing_;
    //! \internal Particle decays in the format {outgoing index, decay products}
    std::unordered_map<size_t, std::vector<Particle>> decays_;

    std::reference_wrapper<HEJ::RNG> ran_;
  };

}
