/** \file
 *  \brief HEJ 2 configuration parameters
 *
 *  \authors   Jeppe Andersen, Tuomas Hapola, Marian Heil, Andreas Maier, Jennifer Smillie
 *  \date      2019
 *  \copyright GPLv2 or later
 */

#pragma once

#include <string>

#include "fastjet/JetDefinition.hh"
#include "yaml-cpp/yaml.h"

#include "HEJ/event_types.hh"
#include "HEJ/HiggsCouplingSettings.hh"
#include "HEJ/optional.hh"
#include "HEJ/output_formats.hh"
#include "HEJ/ScaleFunction.hh"

namespace HEJ{

  //! Jet parameters
  struct JetParameters{
    fastjet::JetDefinition def;          /**< Jet Definition */
    double min_pt;                       /**< Minimum Jet Transverse Momentum */
  };

  //! Settings for scale variation
  struct ScaleConfig{
    //! Base scale choices
    std::vector<ScaleFunction> base;
    //! Factors for multiplicative scale variation
    std::vector<double> factors;
    //! Maximum ratio between renormalisation and factorisation scale
    double max_ratio;
  };

  //! Settings for random number generator
  struct RNGConfig {
    //! Random number generator name
    std::string name;
    //! Optional initial seed
    optional<std::string> seed;
  };

  /**! Possible treatments for fixed-order input events.
   *
   *  The program will decide on how to treat an event based on
   *  the value of this enumeration.
   */
  enum class EventTreatment{
    reweight,                                  /**< Perform resummation */
    keep,                                      /**< Keep the event */
    discard,                                   /**< Discard the event */
  };

  //! Container to store the treatments for various event types
  using EventTreatMap = std::map<event_type::EventType, EventTreatment>;

  /**! Input parameters.
   *
   * This struct handles stores all configuration parameters
   * needed in a HEJ 2 run.
   *
   * \internal To add a new option:
   *           1. Add a member to the Config struct.
   *           2. Inside "src/YAMLreader.cc":
   *              - Add the option name to the "supported" Node in
   *                get_supported_options.
   *              - Initialise the new Config member in to_Config.
   *                The functions set_from_yaml (for mandatory options) and
   *                set_from_yaml_if_defined (non-mandatory) may be helpful.
   *           3. Add a new entry (with short description) to config.yaml
   *           4. Update the user documentation in "doc/Sphinx/"
   */
  struct Config {
    //! Parameters for scale variation
    ScaleConfig scales;
    //! Resummation jet properties
    JetParameters resummation_jets;
    //! Fixed-order jet properties
    JetParameters fixed_order_jets;
    //! Minimum transverse momentum for extremal partons
    double min_extparton_pt;
    //! Maximum transverse momentum fraction from soft radiation in extremal jets
    double max_ext_soft_pt_fraction;
    //! Number of resummation configurations to generate per fixed-order event
    int trials;
    //! Whether to include the logarithmic correction from \f$\alpha_s\f$ running
    bool log_correction;
    //! Event output files names and formats
    std::vector<OutputFile> output;
    //! Parameters for random number generation
    RNGConfig rng;
    //! Map to decide what to do for different event types
    EventTreatMap treat;
    //! Parameters for custom analyses
    YAML::Node analysis_parameters;
    //! Settings for effective Higgs-gluon coupling
    HiggsCouplingSettings Higgs_coupling;
  };

  //! Configuration options for the PhaseSpacePoint class
  struct PhaseSpacePointConfig {
    //! Properties of resummation jets
    JetParameters jet_param;
    //! Minimum transverse momentum for extremal partons
    double min_extparton_pt;
    //! Maximum transverse momentum fraction from soft radiation in extremal jets
    double max_ext_soft_pt_fraction;
  };

  //! Configuration options for the MatrixElement class
  struct MatrixElementConfig {
    //! Whether to include the logarithmic correction from \f$\alpha_s\f$ running
    bool log_correction;
    //! Settings for effective Higgs-gluon coupling
    HiggsCouplingSettings Higgs_coupling;
  };

  //! Configuration options for the EventReweighter class
  struct EventReweighterConfig {
    //! Settings for phase space point generation
    PhaseSpacePointConfig psp_config;
    //! Settings for matrix element calculation
    MatrixElementConfig ME_config;
    //! Properties of resummation jets
    JetParameters jet_param;
    //! Treatment of the various event types
    EventTreatMap treat;
  };

  /**! Extract PhaseSpacePointConfig from Config
   *
   * \internal We do not provide a PhaseSpacePointConfig constructor from Config
   * so that PhaseSpacePointConfig remains an aggregate.
   * This faciliates writing client code (e.g. the HEJ fixed-order generator)
   * that creates a PhaseSpacePointConfig *without* a Config object.
   *
   * @see to_MatrixElementConfig, to_EventReweighterConfig
   */
  inline
  PhaseSpacePointConfig to_PhaseSpacePointConfig(Config const & conf) {
    return {
      conf.resummation_jets,
      conf.min_extparton_pt,
      conf.max_ext_soft_pt_fraction
    };
  }

  /**! Extract MatrixElementConfig from Config
   *
   * @see to_PhaseSpacePointConfig, to_EventReweighterConfig
   */
  inline
  MatrixElementConfig to_MatrixElementConfig(Config const & conf) {
    return {conf.log_correction, conf.Higgs_coupling};
  }

  /**! Extract EventReweighterConfig from Config
   *
   * @see to_PhaseSpacePointConfig, to_MatrixElementConfig
   */
  inline
  EventReweighterConfig to_EventReweighterConfig(Config const & conf) {
    return {
      to_PhaseSpacePointConfig(conf),
      to_MatrixElementConfig(conf),
      conf.resummation_jets, conf.treat
    };
  }

} // namespace HEJ
